exception Error of string

let build e =
  let rec build' e env = match e with
    | Expr.Int x            -> Type.Int, []
    | Expr.Bool b           -> Type.Bool, []
    | Expr.Var id           -> (try Env.find env id, []
				with Not_found -> raise (Error ("Variable '"^id^"' is unbound in current environment.")))
    | Expr.Op (op,l)        -> build_op (op,l) env
(*    | Expr.Let (id,f,f')    -> (let tf,lf = build' f env in
				let tf',lf' = build' f' (Env.add env id tf) in (tf',lf@lf'))
    | Expr.LetRec (id,f,f') -> (let tf0 = Type.new_var () in
				let tf,lf = build' f (Env.add env id tf0) in
				let tf',lf' = build' f' (Env.add env id tf0) in (tf',(tf0,tf)::lf@lf')) *)
    | Expr.Fun (id,f)       -> (let tv = Type.new_var () in
				let tf,lf = build' f (Env.add env id tv) in
				  Type.Fun (tv,tf), lf)
    | Expr.App (f,f')       -> (let tf,lf = build' f env in
				let tv = Type.new_var () and tr = Type.new_var () in
				let tf',lf' = build' f' env in
				  (tr, (tf,Type.Fun (tv,tr))::(tf',tv)::lf@lf'))
    | _                     -> raise (Error "Unsupported construct.")

  and build_op (op,l) env = match (op,l) with
    | (Expr.Add, [x;y])
    | (Expr.Sub, [x;y])
    | (Expr.Mul, [x;y])
    | (Expr.Div, [x;y])  -> (let tx,lx = build' x env and ty,ly = build' y env in
			       (Type.Int, (tx,Type.Int)::(ty,Type.Int)::lx@ly))
    | (Expr.Not, [x])    ->  let tx,lx = build' x env in Type.Bool, (tx,Type.Bool)::lx
    | (Expr.And, [x;y])  -> (let tx,lx = build' x env and ty,ly = build' y env in
			       (Type.Bool, (tx,Type.Bool)::(ty,Type.Bool)::lx@ly))
    | (Expr.Eq, [x;y])   -> (let tx,lx = build' x env and ty,ly = build' y env in
			       (Type.Bool, (tx,ty)::lx@ly))
    | (Expr.Gt, [x;y])   -> (let tx,lx = build' x env and ty,ly = build' y env in
			       (Type.Bool, (tx,Type.Int)::(ty,Type.Int)::lx@ly))
    | (Expr.If, [c;x;y]) -> (let tc,lc = build' c env and tx,lx = build' x env and ty,ly = build' y env in
			       (tx, (tc,Type.Bool)::(tx,ty)::lc@lx@ly))
    | _                  ->  raise (Error "Syntax error.")

  in build' e (Env.create ())
