type t = Cons of string * (t list) | Var of int

exception Error of string

let rec print = function
  | Cons (s,l) -> print_string (s^" ( "); List.iter (fun x -> print x; print_string " ") l; print_string ")"
  | Var i      -> print_string "X"; print_int i

let rec list_var = function
  | Cons (_,l) -> List.flatten (List.map list_var l)
  | Var i      -> [i]

let rec replace x i x' = match x with
  | Cons (s,l) -> Cons (s, List.map (fun x -> replace x i x') l)
  | Var j      -> if j = i then x' else Var j

let unif p =
  let rec aux (p,s) = match p with
    | [] -> s
    | c::q -> (match c with
		 | x, y                       when x  = y  -> aux (q, s)
		 | Cons (sx,lx), Cons (sy,ly) when sx = sy -> aux ((List.map2 (fun x y -> x,y) lx ly)@q, s)
		 | Cons (sx,lx), Cons (sy,ly)              -> raise (Error ("Cannot unify constructors '"^sx^"' and '"^sy^"'."))
		 | Var i, y                                -> (if List.mem i (list_var y) then
								 raise (Error "Cyclic constraint.")
							       else
								 aux (List.map (fun (x',y') -> (replace x' i y,
												replace y' i y)) q,
								      (i,y)::(List.map (fun (j,x') -> (j,replace x' i y)) s)))
		 | x, Var j                                -> aux ((Var j, x)::q, s))
  in aux (p,[])
